// FIXME: no hay ninguna estructura nueva, se utilizan las mismas que en la clase PolarSectors:
// se debe definir todas las estructuras que se definen en la librería PolarSectors.h e incluirse
// todas las librerías que se incluyen en ésta.
// #include "PolarSectors.h"

#include "./CJoseph.h"

CJoseph::CJoseph()
{
		m_SectorMatrix = NULL;
		m_Sector2D = NULL;
		m_VOI_to_2D = NULL;	
}

CJoseph::~CJoseph()
{
	if(m_SectorMatrix != NULL){
		delete m_SectorMatrix;
		m_SectorMatrix = NULL;
	}

	if(m_Sector2D != NULL){
		delete m_Sector2D;
		m_Sector2D = NULL;
	}

	if(m_VOI_to_2D != NULL){
		delete m_VOI_to_2D;
		m_VOI_to_2D = NULL;
	}
}

	// Proyecta un centro dado al plano de y = 0, dado este centro y la altura del foco (m_R)
	double CJoseph::proyectarCentroSobreYCero(VEC2D_double centro)
	{
		VEC2D_double v, focal; // Vector que va del foco al centro que queremos proyecctar

		// Cálculo de las componentes del vector que va del foco al centro que queremos proyectar
		focal.x = 0.0;
		focal.y = m_R;

		v.x = centro.x - focal.x;
		v.y = centro.y - focal.y;

		double lambda = focal.y / v.y;
		return v.x * lambda;
	}

	int CJoseph::Grid_Size(){
		if(m_State != GEOM_DEFINED){
			return IFIC_ERROR;
		}

		// FOV Parameters
		m_alpha = 2.0 * g_PI / (double) m_NProy;
		double AreaDetect = (double) m_NDetect * m_DetectSize / 2.0; // Area en pixels
		m_FOV_Radius = AreaDetect * m_R / sqrt((AreaDetect * AreaDetect) + (m_R * m_R));

		m_NAnillos = (int) (m_FOV_Radius / m_DetectSize);


		FREE_PTR(m_NPixelsPorAnillo);
		m_NPixelsPorAnillo = (int *) calloc(m_NAnillos, sizeof(int));
		if(m_NPixelsPorAnillo == NULL)
			return IFIC_ERROR;

		if(m_RadioPixels != NULL){
			free(m_RadioPixels);
			m_RadioPixels = NULL;
		}
		m_RadioPixels = (double *) calloc(m_NAnillos, sizeof(double));
		if(m_RadioPixels == NULL){ 
			return IFIC_ERROR; }

		m_NPixelsPorSector = 0;
		int NPixels = 0;
		for(int m = 0; m < m_NAnillos; m++){
			NPixels = (int) (2.0 * g_PI * (m + 0.5) / (double) m_NProy);
			NPixels++;

			m_NPixelsPorAnillo[m] = NPixels;
			m_NPixelsPorSector += NPixels;
			m_RadioPixels[m] = (2.0 * g_PI / (double) (m_NProy * NPixels));
		}
		m_NCorners = 2 * (m_NPixelsPorSector * m_NAnillos);

		return IFIC_OK;
	}

	int CJoseph::Genera_Posiciones()
	{
		if(PreparePositionMatrices() != IFIC_OK){
			return IFIC_ERROR;
		}

		GridPos * Voxel = m_Grid;
		GridPos * RotVoxel = m_RotatedGrid;
		VEC2D_double * Corner = m_Corners;
		VEC2D_double * RotCorner = m_RotatedCorners;
		VEC2D_double * centro	 = m_Centres;


		// El primer anillo tiene un cálculo especial proque sólo tiene tres puntos
		double area, angulo, anguloCentro, centroAnillo;
		for(int m = 0; m < m_NAnillos; m++){
			area = (m+0.5) * m_RadioPixels[m];

			// Cálculo del punto central del anillo
			// TODO: Averiguar donde rotar los centros calculados
			// se supone un anillo por detector
			centroAnillo = (double) (m + 0.5)* m_DetectSize;

			for(int k = 0; k < m_NPixelsPorAnillo[m]; k++, Voxel++, RotVoxel++){
				Voxel->r = area;
				Voxel->Pos = Corner;
				RotVoxel->r = area;
				RotVoxel->Pos = RotCorner;
				RotCorner += 2;

				angulo = (double) k * m_RadioPixels[m];
				anguloCentro = (double) (k+0.5) * m_RadioPixels[m];

				Corner->x = (double) (m * m_DetectSize * cos(angulo));
				Corner->y = (double) (m * m_DetectSize * sin(angulo));
				Corner++;
				Corner->x = (double) ((m+1) * m_DetectSize * cos(angulo));
				Corner->y = (double) ((m+1) * m_DetectSize * sin(angulo));
				Corner++;

				// Cálculo del centro del pixel
				// FIXME: es posible que el factor de la x sea cos y el de la y -sin ?
				centro->x = centroAnillo * cos(anguloCentro);
				centro->y = centroAnillo * sin(anguloCentro);
				centro++;
			}
			// FIXME: este calculo se puede incuir dentro del bucle si la guarda es <= y no <
			// Cálculo de las dos esquinas finales de cada anillo
			angulo = (double) m_NPixelsPorAnillo[m] * m_RadioPixels[m];
			Corner->x = (double) (m * m_DetectSize * cos(angulo));
			Corner->y = (double) (m * m_DetectSize * sin(angulo));
			Corner++;
			Corner->x = (double) ((m+1) * m_DetectSize * cos(angulo));
			Corner->y = (double) ((m+1) * m_DetectSize * sin(angulo));
			Corner++;
			RotCorner += 2;
		}

		memcpy(m_RotatedCorners, m_Corners, m_NCorners * sizeof(VEC2D_double));
		return IFIC_OK;
	}

	int CJoseph::Prepare_Weigths_Calculation()
	{
		if(m_SectorMatrix != NULL){
			delete m_SectorMatrix;
			m_SectorMatrix = NULL;
		}
		m_SectorMatrix = new CSectorLOR;
		if(m_SectorMatrix == NULL){ return IFIC_ERROR; }

		BackProject_Mtx * LOR_Projection = new BackProject_Mtx;
		LOR_Projection->NElements = m_tmp_length;
		LOR_Projection->LOR = m_tmp_LOR;
		LOR_Projection->Weight = m_tmp_Weight;
		LOR_Projection->Pixel = m_tmp_idx;

		if(m_SectorMatrix->InitPointers(m_LorVectors, m_NDetect, m_DetectSize, m_R, m_LorHistogram, LOR_Projection) != IFIC_OK){
			delete LOR_Projection;
			return IFIC_ERROR;
		}
		delete LOR_Projection;
		return IFIC_OK;
	}

	int CJoseph::Calculate_SectorWeights(double angle, int *NElements){
		// Comprobación de que la estructura que va a generar los pesos está preparada
		if(Rotate_Grid(angle) != IFIC_OK){
			return IFIC_ERROR;
		}
		GridPos * PixelPos = m_RotatedGrid;

		if(m_SectorMatrix->InitSector() != IFIC_OK){
			return IFIC_ERROR;
		}

		// TODO: Implementar Joseph
		int NDetectDiv2 = m_NDetect / 2;
		double proyeccion;
		int elemento = 0;
		int indiceLORAnterior, indiceLORPosterior;
		double pesoLORAnterior, pesoLORPosterior;
		for(int m = 0; m < m_NPixelsPorSector; m++){
			// Obtención de la proyección del centro en el plano Y = 0
			proyeccion = proyectarCentroSobreYCero(m_RotatedCentres[m]);
			// Normalización respecto al tamaño del detector
			proyeccion /= m_DetectSize;
			// Cálculo del indice de la LOR anterior
			indiceLORAnterior = (int) floor(proyeccion);
			indiceLORAnterior += NDetectDiv2;
			// Cálculo del indice de la LOR posterior
			indiceLORPosterior = indiceLORAnterior + 1;
			// Cálculo de la contribución del pixel a la LOR anterior
			pesoLORAnterior = indiceLORPosterior - (proyeccion + NDetectDiv2);
			// Cálculo de la contribución del pixel a la LOR posterior
			pesoLORPosterior = (proyeccion + NDetectDiv2) - indiceLORAnterior;

			// Almacenamiento de los pesos
			// FIXME: se esta guardadno correctamente? es la variable correcta? el orden correcto?
			// Almacenamiento de la LOR anterior al pixel 
			m_tmp_LOR[elemento] = indiceLORAnterior;
			m_tmp_idx[elemento] = m;
			m_tmp_Weight[elemento] = pesoLORAnterior;
			elemento++;
			// Almacenamiento de la LOR posterior al pixel 
			m_tmp_LOR[elemento] = indiceLORPosterior;
			m_tmp_idx[elemento] = m;
			m_tmp_Weight[elemento] = pesoLORPosterior;
			elemento++;

			// Pureba de que todo ha ido bien
			if( !((indiceLORAnterior < (proyeccion + NDetectDiv2)) &&
				((proyeccion + NDetectDiv2) < indiceLORPosterior) &&
				((pesoLORAnterior > 0) && (pesoLORPosterior > 0)) &&
				((pesoLORAnterior < 1) && (pesoLORPosterior < 1))) ){
					return IFIC_ERROR;
			}
		}
		m_tmp_length = elemento;

		// FIXME: (posible error) en el código original de PolarSectors.cpp aquí se devuelve IFIC_OK
		if(m_SectorMatrix->GetNElemnts(NElements) != IFIC_OK){
			return IFIC_ERROR;
		}
		return IFIC_OK;
	}
	

	int CJoseph::Extrae_Corners(GridPos *Pixel){
		VEC2D_double * Corner = Pixel -> Pos;

		m_VoxelCorners[0].x = Corner[0].x;
		m_VoxelCorners[0].y = Corner[0].y;

		m_VoxelCorners[1].x = Corner[1].x;
		m_VoxelCorners[1].y = Corner[1].y;

		// FIXME: los indices están cambiados a propósito??

		m_VoxelCorners[2].x = Corner[3].x;
		m_VoxelCorners[2].y = Corner[3].y;

		m_VoxelCorners[3].x = Corner[2].x;
		m_VoxelCorners[3].y = Corner[2].y;

		return IFIC_OK;
	}

	int CJoseph::Polar2Cartesian(int ** Cartesian, int ** Polar, double ** Weight, int * NWeights){
		if(Prepare_2D_Proj() != IFIC_OK){
			return IFIC_ERROR;
		}
		if(Calculate_2D_Proj() != IFIC_OK){
			return IFIC_ERROR;
		}
		if(Ordena_2D_Proj() != IFIC_OK){
			return IFIC_ERROR;
		}
		/**
		* if(Normalize_2D_Proj() != IFIC_OK){
		* 	return IFIC_ERROR;
		* }
		*/
		// Copia de punteros
		*NWeights = m_P2Cart_NIdx;
		*Weight = m_P2Cart_Weights;
		*Polar = m_Polar_Idx;
		*Cartesian = m_Cart_Idx;

		return IFIC_OK;
	}

	int CJoseph::Ordena_2D_Proj(){
		m_P2Cart_NIdx = 0;
		int NProy = m_NProy;

		// Cálculo del número de elementos
		for(int m = 0; m < NProy; m++){
			m_P2Cart_NIdx += m_VOI_to_2D[m].NElements;
		}

		// Reserva de memoria para las matrices que guardan la proyección
		if(m_P2Cart_Weights != NULL){
			free(m_P2Cart_Weights);
			m_P2Cart_Weights = NULL;
		}
		m_P2Cart_Weights = (double *) calloc(m_P2Cart_NIdx, sizeof(double));
		if(m_P2Cart_Weights == NULL){ return IFIC_ERROR; }

		if(m_Polar_Idx != NULL){
			free(m_Polar_Idx);
			m_Polar_Idx = NULL;
		}
		m_Polar_Idx = (int *) calloc(m_P2Cart_NIdx, sizeof(int));
		if(m_Polar_Idx == NULL){ return IFIC_ERROR; }

		if(m_Cart_Idx != NULL){
			free(m_Cart_Idx);
			m_Cart_Idx = NULL;
		}
		m_Cart_Idx = (int *) calloc(m_P2Cart_NIdx, sizeof(int));
		if(m_Cart_Idx == NULL){ return IFIC_ERROR; }

		int *Cart = m_Cart_Idx;
		int *Polar = m_Polar_Idx;
		double *Weight = m_P2Cart_Weights;
		int NElements;
		for(int m = 0; m < NProy; m++){
			NElements = m_VOI_to_2D[m].NElements;
			memcpy(Cart,	 m_VOI_to_2D[m].LOR, 	NElements*sizeof(int));
			memcpy(Polar,	 m_VOI_to_2D[m].Pixel, 	NElements*sizeof(int));
			memcpy(Weight,	 m_VOI_to_2D[m].Weight, NElements*sizeof(int));
			Cart 	+= NElements;
			Polar 	+= NElements;
			Weight 	+= NElements;
		}
		return IFIC_OK;
	}

	int CJoseph::Normalize_2D_Proj(){
		float * Suma_Mtx;
		int NElem;

		if(m_Sector2D->GetNormalizationMtx(&Suma_Mtx, &NElem) != IFIC_OK){
			return IFIC_ERROR;
		}

		float * Normaliza = (float *) calloc(NElem, sizeof(float));
		if(Normaliza == NULL){ return IFIC_ERROR; }

		for(int m = 0; m < NElem; m++){
			Normaliza[m] = (double) (Suma_Mtx[m] == 0.0) ? 0.0 : (1.0 / Suma_Mtx[m]);
		}

		double * Pesos = m_P2Cart_Weights;
		int * Cart = m_Cart_Idx;
		for(int m = 0; m < m_P2Cart_NIdx; m++, Cart++, Pesos++){
			*Pesos *= Normaliza[*Cart];
		}
		free(Normaliza);
		return IFIC_OK;
	}

	int CJoseph::Prepare_2D_Proj(){
		if(Prepara_TmpMtx(8) != IFIC_OK){
			return IFIC_ERROR;
		}
		if(m_VOI_to_2D != NULL){
			delete m_VOI_to_2D;
			m_VOI_to_2D = NULL;
		}
		m_VOI_to_2D = new BackProject_Mtx[m_NProy];
		// FIXME: (posible error) en el código original de PolarSectors.cpp aquí se devuelve IFIC_OK
		if(m_VOI_to_2D == NULL){ return IFIC_ERROR; }

		if(m_Sector2D != NULL){
			delete m_Sector2D;
			m_Sector2D = NULL;
		}
		m_Sector2D = new CSector2D;
		if(m_Sector2D == NULL){ return IFIC_ERROR; }

		BackProject_Mtx *tmpMtx = new BackProject_Mtx;
		tmpMtx->NElements = m_tmp_length;
		tmpMtx->LOR = m_tmp_LOR;
		tmpMtx->Pixel = m_tmp_idx;
		tmpMtx->Weight = m_tmp_Weight;

		// FIXME: (posible error) La división y la multiplicación no tienen paréntesis:
		//                         el orden de preferencia de las operaciones se deja
		//                         dependiente del compilador que no necesáriamente las 
		//                         ejecutará de izquierda a derecha.
		double Vmin = -(double)m_NDetect / 2.0 * m_DetectSize;
		double Hmin = -(double)m_NDetect / 2.0 * m_DetectSize;

		int InitPtr = m_Sector2D->InitPointers(m_NDetect, Vmin, m_NDetect, Hmin, m_DetectSize, tmpMtx);

		delete tmpMtx;
		return InitPtr;
	}

	int CJoseph::Calculate_2D_Proj()
	{
		int NElemInSector;
		GridPos *VoxelPos;
		int Voxel;

		for(int m = 0; m < m_NProy; m++){
			if(m_Sector2D->InitSector() != IFIC_OK){
				return IFIC_ERROR;
			}

			// Rotación de la matriz de pesos
			Rotate_Grid(m * m_alpha);

			VoxelPos = m_RotatedGrid;
			Voxel = m * m_NPixelsPorSector;

			for(int k = 0; k < m_NPixelsPorSector; k++, VoxelPos++){
				if(Extrae_Corners(VoxelPos) != IFIC_OK){
					return IFIC_ERROR;
				}
				if(m_Sector2D->SectoralVoxel_Projection(4, m_VoxelCorners, Voxel + k) != IFIC_OK){
					return IFIC_ERROR;
				}
			}
			m_Sector2D->GetNElemnts(&NElemInSector);
			// Guarda memoria
			m_VOI_to_2D[m].NElements = NElemInSector;

			m_VOI_to_2D[m].LOR = (int *) calloc(NElemInSector, sizeof(int));
			if(m_VOI_to_2D[m].LOR == NULL){ return IFIC_ERROR; }

			m_VOI_to_2D[m].Pixel = (int *) calloc(NElemInSector, sizeof(int));
			if(m_VOI_to_2D[m].Pixel == NULL){ return IFIC_ERROR; }

			m_VOI_to_2D[m].Weight = (double *) calloc(NElemInSector, sizeof(double));
			if(m_VOI_to_2D[m].Weight == NULL){ return IFIC_ERROR; }

			memcpy(m_VOI_to_2D[m].LOR, 		m_tmp_LOR, 		NElemInSector * sizeof(int));
			memcpy(m_VOI_to_2D[m].Pixel, 	m_tmp_idx, 		NElemInSector * sizeof(int));
			memcpy(m_VOI_to_2D[m].Weight, 	m_tmp_Weight, 	NElemInSector * sizeof(double));
		}
		return IFIC_OK;
	}
